/*
   output.js $Id: output.js,v 1.69 2002/09/03 00:28:52 johnh Exp $
   =========

   Copyright (c) John Holdsworth, Coldwave Programs Ltd. 2001
   This code is distributed for evaluation only without the right
   to redistribute with no warantees either express or implied.

   This class performs grouping operations on a set of XML
   nodes and renders the nodes into HTML. It should be used
   in conjunction with a user interface controller class
   which maintains the groupings and columns the user wishes
   to view (e.g. xtab/gui.js).

   API:

   xmlDoc = newXML( xmlString )
	A central function to create a new instance of an XML document 
	from a string of XML. The version of MSXML used by all code
	is selected in this function (currently 3.0 sp1)

   columnName = defix( fullColumnName )
	Removes any prefixes from a column name to make it
	suitable for display as a column heading. The prefixes
	defined at the start of a column name are currently:

	# - do not display this column initialy. Useful for hiding
	    underlying low level keys but having them available for
	    joins

  	= - Group and total by this column in the initial display

	+ - Add an initial asceding sort by this column.

	- - Add an initial descending sort by this column.

	nn.nn - control the initial width of a column (not implemented)

    xtab = XTab( xmlString )
	Constructor of a interactive xml data document for the string of XML.
	Data islands are no longer used as they create a pre-version 3
	instance of an xmlDoc and have to be converted back into a
	string and reparsed before they can be used.

	The required node structure  is:

	<grp>
	  <column name='#' type='count' attr='c0'/>
	  <column name='columnName' type='text|float|int'/>
	  <column name='columnName' type='text|float|int'/>
          ...
	  <row c0='1' c1='column1' c2='column2 .../>
	  <row c0='2' c1='column1' c2='column2 .../>
	  ...
	</grp>

	The column nodes are parsed and represented in the
	this.columns array instance varaiable then removed.

     groupBy( level, rootNode, keycols [], grouper )
	Perform a grouping of row records contained in the
	node acording to a key value which is obtained by
	concatenating values in the columns named in the
	keycols argument with a "~" character. Any previous
	grouping is replaced.
	
	This function is rather over-optimised for speed
	and synthesises the code to scan each row node for
	its key at the last minute. The level argument
	specifies which level of grouping is being specified
	as groupings can be performed one after the other.
	A level one grouping means: for each grouping
	sub-node of the root node passed in perform a
	sub-grouping using the keycols specified.

	After grouping the document nodes are arranged
	for example grouping by column "c1":

	<grp>
	  <grp c1="a1">
	    <row c1="a1" c2="v2"
	    <row c1="a1" c2="u2"
	  </grp>
	  <grp c1="b1">
	    <row c1="b1" c2="v2"
	    <row c1="b1" c2="u2"
	  </grp>
	</grp>

      doGroups( groups[], rootNode )
	Called to perform a scuession of groupings and
	subgroupings on a node. For example calling:

	doGroups( [["x"], ["y,z"]], xtab.data.documentElement )

	Will rearrange the nodes in the XML document
	to be grouped first by the values of column x
	then by the combined value of columns "y" and "z"
	concatenated with a "~".

      unGroup( rootNode )
	Remove all grouping sub nodes from the node specified.

      xslt = buildFormatter( columns, sorts )
	Builds an xslt transformation to render an XML
	document of grouped record nodes into HTML.

	This is performed in the following steps:

	The columns specified are looked up in the columns
	array and their type and column attribute name
	determined.

	For each column the type is used to construct an
	xslt fragment to render a <td> cell for the column.
	Float and integer values sum across any grouping,
	text values are displayed only if they have been used
	in a higher level grouping or there is only one record.

	These fragments are substituted along with any sorts
	into the text of a template xslt transformation using
	a regular explression. This transformation can then
	be applied to a node tree to create an HTML table element.

	Due to the way the templates are applied sorts are
	applied at each level of grouping, also if a <grp>
	or any of its ancestors contain an attribute "_open"
	the template will recurse into the row records and
	render them.

	The style class for cells in a row is determined
	by the depth of grouping (number of ancestors)
	and can be changed by tweaking the style sheet.

      render( cols, sorts )
	builds a xslt formatter and either document.write()'s
	the HTML to the user's page first time around or 
	substitues it as the "innerHTML" of the <div> element
	containing the table if it has already been rendered.

*/

//**Start Encode**

// XML doc factory
function newXML ( source ) {
    var root = new ActiveXObject( "MSXML2.DOMDocument.3.0" );
    root.async = false;
    var sstr = ""+source;

    (sstr.match( /\[object\]/ ) ? root.loadXML( source.xml ) :
	sstr.match( /^http:\// ) ? root.load( window.dataURL = source ) :
	root.loadXML( source.replace( /`/g, "'" ) )
    ) || alert( "Could not load XML:\n"+source );

    return root;
}

// normalise column names (remove sorts, precicion etc)
// name can contain =[+-]#.nnn as described above.
function defix (name) {
    return name.length > 1 ? name.replace( /(^|,)[-+=#\d.$]+/g, "$1" ) : name;
}

// grouping nodes are given unique negative node numbers
var seq = 0;

// contructor for new Xtab instance
function XTab (island, aGui) {
   this.gui = aGui;
   this.rname = "rendered"+aGui.gno;

  // try to create use C++ Xtab implementation
//  try { 
//    this.xmlc = new ActiveXObject( "Xtab.XtabControl" );
//    this.data = this.island = this.xmlc.docRoot;
//    this.data.async = false;
//    (island.match( /^http:/ ) ? this.data.load( window.dataURL=island ) : 
//	this.data.loadXML( island ))||alert( "Could not load XML:\n"+source );
//  }
//  catch( e ) {
    // use normal MSXML3 instance
    this.data = this.island = newXML( island );
//  }

  // select of column Meta-data out of xml recordset
  var colNodes = this.data.selectNodes( "/*/column" );
  var columns = this.columns = [];
  var pKeys = this.pKeys = {};
  var ptab, pcol;
  
  // default table name for update
  if ( window.sql && window.sql.match( /from\s+(\w+)\s+--\s+(\w+)/ ) ) {
	ptab = RegExp.$1;
	pcol = RegExp.$2;
  }

  // build column mappings and initial attributes
  for ( var c=0 ; c<colNodes.length ; c++ ) {
    columns[c] = { attr:"c"+c };

    // copy element attributes into JS Object
    var att = colNodes[c].attributes;
    for ( var i =0; i<att.length ; i++ ) {
	columns[c][att[i].nodeName] = att[i].nodeValue;

        // set up primary key info for updates
	if ( att[i].nodeName == "pkey" ) {
	  var tab = colNodes[c].getAttribute("tab");
	  if ( !pKeys[tab] )
	    pKeys[tab] = [];
	  pKeys[tab][pKeys[tab].length] = columns[c];
	}

    }

    //  build map from real to internal col name
    var name = columns[c].name;
    columns[defix(name)] = columns[c];

    // extract precision parameter
    if ( len = name.substring( 1 ).match( /\d+/ ) )
	columns[c].length = len;

    // default column source table
    if ( !columns[c].tab  ) {
	columns[c].tab = ptab;
	columns[c].col = defix(name);
    }

    // insert default value & remove column record from from dataset
    // this.data.documentElement.setAttribute( columns[c].attr, "&nbsp;" );
    this.data.documentElement.removeChild( colNodes[c] );
  }

  // fill in default primary key
  var kcol = columns[pcol];
  if ( kcol )  {
      pKeys[ptab] = [kcol];
      kcol.pkey = 1;
  }
}

// template routine for fast grouping
// compiled when keys are known for speed
var groupcode = '\
  var row, groupNode, groups = {};\
\
  while( row = rows.nextNode() ) {\
   if ( !(groupNode = groups[key = __KEYEXPR__]) ) {\
	groupNode = groups[key] = this.data.documentElement.cloneNode(false);\
\
__ATTRSETS__\
\
        groupNode.setAttribute( "c0", "g"+-(--seq) );\
	root.appendChild( groupNode );\
    }\
\
    groupNode.appendChild( row );\
  }';

// core multi-key recurive grouping routine
function XTab.prototype.groupBy ( level, root, keys, grouper ) {

  // if not being called recursively..
  if ( !grouper ) {
    // build grouping routine mapping
    // external to internal column names
    var keyExpr = "", attrSets = "";
    var cols = keys.match( /[^,]+/g )||[];
    keys = [];

    for ( var k=0 ; k<cols.length ; k++ ) {
      keys[k] = this.columns[cols[k]].attr;
      keyExpr += (!k?"":"+\"~\"+")+"row.getAttribute( \""+keys[k]+"\" )";
      attrSets += "\tgroupNode.setAttribute( '"+keys[k]+
	"', row.getAttribute( '"+keys[k]+"' ) );\n";
    }

    // build grouper and create function for this key set
    code = groupcode.replace( /__KEYEXPR__/, keyExpr || '""' )
	.replace( /__ATTRSETS__/, attrSets );
    grouper = new Function( "rows, root, keys", code );
  }

  // list existing grouping nodes at this level.
  var gnodes = root.selectNodes( "grp" );
  gnodes.length; // cache node list..

  if ( level-- ) {
    // recurse if this is not the node level
    // that this set of keys applies to
    var gnode;

    while( gnode = gnodes.nextNode() )
      this.groupBy( level, gnode, keys, grouper );

    return;
  }

  // if final key set is empty then display all records
  root.setAttribute( "_open", keys.length ? 0 : 1 );
  if ( !keys.length )
    return;

  // ActiveX speedup in C++ for first level groupings
  if ( this.xmlc && root==this.data.documentElement ) {
	for ( var i=0 ; i<keys.length ; i++ )
	    this.xmlc.addKey(  keys[i] );
	this.xmlc.fastGroups( this.data );
	return;
  }

  // otherwise do it in JavaScript using compiled grouper
  var rows = root.selectNodes( ".//row" ); rows.length;
  var row, key, groupNode, groups = {};
  var newRoot = root || this.data.documentElement.cloneNode(false);
	//root || createElement( "grp" );

  this.grouper = grouper;
  this.grouper( rows, newRoot, keys );

  if ( newRoot != root )
    return root.parentNode.replaceChild( newRoot, root );
  else
    while( node = gnodes.nextNode() )
      root.removeChild( node );
}

// grouping entry point, applies a set of levels of groupings
function XTab.prototype.doGroups( groups, root ) {
  since( "gstart" );

  if ( !root || root.tagName == "row" )
    root = this.data.documentElement;

  for ( var g=0 ; g<groups.length ; g++ )
      this.groupBy( g, root, groups[g] );

  since( "gended" );
}

// resets all groupings on this node
function XTab.prototype.unGroup ( doc ) {
  if ( !doc )
    doc = this.data.documentElement;

  var nodes = doc.selectNodes( ".//row" ); nodes.length;
  var node, root = this.data.documentElement.cloneNode(false);

  while( node = nodes.nextNode() )
    root.appendChild( node );

  doc.parentNode.replaceChild( root, doc );
}

//XSLT fragments for formatting primitive types
var formatters = {

	// rowcount type
	count:	'<td><xsl:value-of select="$nrows"/></td>',

	// unrestricted text value
  	text:	'<xsl:choose><xsl:when test="$nrows=1">\
		<xsl:value-of disable-output-escaping="yes"\
	  	select="concat($td,.//__ATTR__)"/>\
		</xsl:when><xsl:otherwise>\
		<xsl:value-of disable-output-escaping="yes"\
		select="concat($td,$ans/__ATTR__)"/>\
		</xsl:otherwise></xsl:choose>',

	// text with width restriction (precision)
  	textn:	'<xsl:choose><xsl:when test="$nrows=1">\
		<xsl:value-of disable-output-escaping="yes"\
		select="concat($td,substring(\
		.//__ATTR__,1,__LEN__))"/></xsl:when><xsl:otherwise>\
		<xsl:value-of disable-output-escaping="yes"\
		select="concat($td,substring(\
		$ans/__ATTR__,1,__LEN__))"/>\
		</xsl:otherwise></xsl:choose>',

	// integer value (summed across row(s))
	int:	'<xsl:value-of disable-output-escaping="yes"\
		select="concat(\'&lt;td class=int&gt;\',\
		sum($rows//__ATTR__))"/>',

	// floating (money) values, commas inserted
	// with negative values displayed in red
	float: '<xsl:element name="td">\
		<xsl:variable name="class" select="$rows/__ATTR___"/>\
		<xsl:variable name="value" select="sum($rows/__ATTR__)"/>\
		<xsl:attribute name="class"><xsl:choose>\
		<xsl:when test="$class"><xsl:value-of select="$class"/>\
		</xsl:when><xsl:otherwise><xsl:choose>\
		<xsl:when test="$value &lt; 0">neg</xsl:when>\
		<xsl:otherwise>pos</xsl:otherwise></xsl:choose>\
		</xsl:otherwise></xsl:choose></xsl:attribute>\
		<xsl:value-of select="format-number($value, \',##0.00\')"/>\
		</xsl:element>',


	// not implemented
	link: 	'<td><a href="a"><xsl:value-of select=".//__ATTR__"/></a></td>'
};

// XSLT tranformation from XML doc to HTML table element
// recurses when there are subgroupings or the node is "_open"
// XSLT fragments above are substituted in to build final formatter.
//	<?xml version="1.0" encoding="utf-8"?>
// 	<xsl:output method="html" encoding="utf-8" omit-xml-declaration="yes"/>
//

var formatter = '\
	<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version = "1.0">\
\
	<xsl:template match="/">\
	 <table border="1" cellspacing="1"\
		 bordercolorlight="black" bordercolordark="blue">\
	<tr id="heading" class="heading" bgcolor="darkblue"\
	title="right click for operations"><HEADINGS/></tr>\
	<xsl:apply-templates select="grp"/>\
	</table></xsl:template>\
\
	<xsl:template match="*">\
	  <xsl:variable name="td" select="\'&lt;td&gt;\'"/>\
	  <xsl:variable name="ans" select="ancestor-or-self::*"/>\
	  <xsl:variable name="rows" select="descendant-or-self::row"/>\
	  <xsl:variable name="nrows" select="count($rows)"/>\
\
	  <xsl:if test="count(grp) or @_open=1">\
	    <xsl:for-each select="*"><SORTS/>\
	      <xsl:apply-templates select="."/></xsl:for-each>\
	 </xsl:if>\
\
	  <xsl:variable name="level" select="count($ans)-1"/>\
\
	  <xsl:element name="tr"><xsl:attribute name="class">\
	    <xsl:choose><xsl:when test="name()=\'row\'">\
	    <xsl:choose><xsl:when test="@_found=1">\
	    found</xsl:when><xsl:otherwise>open</xsl:otherwise></xsl:choose>\
	    </xsl:when><xsl:otherwise>level<xsl:value-of select="$level"/>\
	    </xsl:otherwise></xsl:choose></xsl:attribute>\
	    <xsl:attribute name="_level"><xsl:value-of select="$level"/></xsl:attribute>\
	    <xsl:attribute name="c0"><xsl:value-of select="@c0"/></xsl:attribute>\
	<COLUMNS/>\
	    </xsl:element>\
	  </xsl:template>\
</xsl:stylesheet>';

// builds an XML doc containing an XSLT transformation
// depending on the columns and sorts passed in
function XTab.prototype.buildFormatter ( cols, sorts ) {
  // alert( cols+" -- "+sorts );

  sorts = (sorts||"").match( /([-+=\d$]*[^,]+)/g )||[];
  cols = cols.match( /[-+=\d$]*([^,]+)/g );

  var headings = "";
  var colxsl = "";

  // loop through displayed columns building fomatters
  for ( var c=0 ; c<cols.length ; c++ ) {
    var col = this.columns[cols[c]];
    var heading = col.length ? col.name.substring( 0, col.length ) : col.name;
/*
      .replace(/_/," ");
    if ( col.ftable )
      heading = "<u>"+heading+"</u>";
*/
    headings += "<th class='headcell' title='"+col.dbtype+
	"' colname='"+col.name+"'>"+heading+"</th>";
    colxsl += formatters[col.length? "textn" : col.type]
	.replace( /__ATTR__/g, "@"+col.attr )
	.replace(/__LEN__/g,col.length)+"\n";
  }

  var sortxsl = "";

  // build sort XSLT dependant on type and direction
  for ( var s=0 ; s<sorts.length ; s++ ) {
    var name = ""+sorts[s];
    var col = this.columns[defix(name)];

    sortxsl += "<xsl:sort select=\""+
	( col.type != "text" ? "sum" : "")+"("+".//@"+col.attr+")\""+
	( name.match( /^-/ ) ? " order=\"descending\"" : "" )+
	( col.type != "text" ? " data-type=\"number\"" : "" )+"/>\n";
  }

  // substitute into XSLT source and parse into document
  var xslt = this.data.cloneNode( false );
  xslt.loadXML( formatter
	.replace( /id="heading"/, 'id="heading'+this.gui.gno+'"' )
	.replace( /<HEADINGS\/>/, headings )
	.replace( /<COLUMNS\/>/, colxsl )
	.replace( /<SORTS\/>/g, sortxsl ) );
  return xslt;
}	

function XTab.prototype.render( cols, sorts, dest ) { with (this.gui) {
  var xsl = this.buildFormatter( cols, sorts );
  var html = this.data.transformNode( xsl ).replace( /`/g, "'" )
	.replace( /\t/g, "&nbsp;" ).replace( /-/g, "&minus;" );

  since( "transformed "+html.length );

  if ( !document.all[this.rname] && !dest ) {
    // first time through document.write() is faster
    document.write( '<div ID="'+this.rname+'">' );
    document.write( html );
    document.write( "</div>" );
  }
  else {
    // when page is already rendered, substitute HTML
    if ( !dest ) document.all[this.rname].innerHTML = html;
    if ( dest || this.ldest ) {
        if ( !this.ldest ) { this.ldest = dest; this.lcols = cols };	
        var xsl = this.buildFormatter( this.lcols, sorts );
	this.ldest.innerHTML = this.data.transformNode( xsl )
	   .replace( /\t/g, "&nbsp;" ).replace( /-/g, "&minus;" )
	   .replace( /<td><\//g, "<td>&nbsp;</" );
     }
  }

  this.heading = window.document.all["heading"+this.gui.gno];
  since( "written" );
} }

